from typing import Dict, List, Set
import os
import re
import json

from extr.entities.viewers import HtmlViewer
from ..workspace import load_config, WORKSPACE
from ...utils.filesystem import load_data, save_data, save_document, load_document


def create_redacted_file(annotations: List[str]) -> None:
    def get_redacted_templates() -> Set[str]:
        redacted_templates = set()
        redacted_templates_file_path = os.path.join(WORKSPACE, '4', 'ents-redacted.txt')
        if os.path.exists(redacted_templates_file_path):
            redacted_templates = set(load_data(redacted_templates_file_path))

        return redacted_templates

    redactions: List[str] = [
        re.sub(' +', ' ', re.sub(r'<[A-Z]+>.+?</[A-Z]+>', ' ', row))
        for row in annotations
    ]

    config = load_config()
    if bool(config['annotations']['filter-redactions']):
        redacted_templates = get_redacted_templates()
        redactions = [row for row in redactions if not row in redacted_templates]

    save_data(
        os.path.join(WORKSPACE, '3', 'dev-ents-redacted.txt'),
        redactions
    )

def create_parsed_by_file(text_by_label: Dict[str, List[str]]) -> None:
    save_document(
        os.path.join(WORKSPACE, '3', 'dev-ents.stats.json'),
        json.dumps(text_by_label, indent=2)
    )

def create_html_file(annotations: List[str]) -> None:
    custom_styles = load_document(
        os.path.join(WORKSPACE, 'styles.css')
    )

    save_document(
        os.path.join(WORKSPACE, '3', 'dev-ents.html'),
        HtmlViewer(annotations=annotations).create_view(custom_styles)
    )



from typing import Dict, List
import re
import os
import json

from extr.relations.viewers import HtmlViewer
from extr_ds.models import RelationLabel

from ..workspace import WORKSPACE
from ...utils.filesystem import load_document, save_document


HTML_PATH = os.path.join(WORKSPACE, '3', 'dev-rels.html')
DEV_PATH = os.path.join(WORKSPACE, '3', 'dev-rels.json')

def create_dev_files(relation_groups: Dict[str, List[RelationLabel]]) -> None:
    viewer = HtmlViewer(show_index=True)
    ordered_relations: List[RelationLabel] = []
    for key, items in relation_groups.items():
        viewer.append_header(header=key)
        for relation_label in items:
            viewer.append_relation(
                text=relation_label.original_sentence,
                relation=relation_label.relation
            )

            ordered_relations.append(relation_label)

    save_document(
        DEV_PATH,
        json.dumps(
            list(map(lambda relation: relation.todict(), ordered_relations)),
            indent=2
        )
    )

    save_document(
        HTML_PATH,
        viewer.create_view()
    )

def change_label(label: str, rows: List[int]) -> None:
    html = load_document(HTML_PATH)
    for row in rows:
        html = re.sub(
            r'(<tr id="' + str(row) + '"><td>' + str(row) + '</td><td class="label">)(.+?)(</td>)',
            r'\1' + label + r'\3',
            html
        )

    save_document(HTML_PATH, html)

    dev = json.loads(load_document(DEV_PATH))
    for row in rows:
        dev[row] = {
            'sentence': dev[row]['sentence'],
            'label': label,
            'definition': dev[row]['definition'],
        }

    save_document(DEV_PATH, json.dumps(dev, indent=2))

def delete_row(rows: List[int]) -> None:
    html = load_document(HTML_PATH)
    for row in rows:
        html = re.sub(
            r'(<tr )(id="' + str(row) + '")',
            r'\1class="delete" \2',
            html
        )

    save_document(HTML_PATH, html)

    dev = json.loads(load_document(DEV_PATH))
    for row in rows:
        dev[row] = {
            'sentence': dev[row]['sentence'],
            'label': dev[row]['label'],
            'definition': dev[row]['definition'],
            'attribute': 'delete',
        }

    save_document(DEV_PATH, json.dumps(dev, indent=2))

def recover_row(rows: List[int]) -> None:
    html = load_document(HTML_PATH)
    for row in rows:
        html = re.sub(
            r'(<tr) class="delete" (id="' + str(row) + '")',
            r'\1 \2',
            html
        )

    save_document(HTML_PATH, html)

    dev = json.loads(load_document(DEV_PATH))
    for row in rows:
        dev[row] = {
            'sentence': dev[row]['sentence'],
            'label': dev[row]['label'],
            'definition': dev[row]['definition'],
        }

    save_document(DEV_PATH, json.dumps(dev, indent=2))



from typing import List, Set, Union

import os


def load_document(file_path: str, default_value = '') -> str:
    document = default_value
    if os.path.exists(file_path):
        with open(file_path, 'r', encoding='utf-8') as input_file:
            document = input_file.read()

    return document

def save_document(file_path: str, document: str) -> None:
    with open(file_path, 'w', encoding='utf-8') as output:
        output.write(document)

def load_data(file_path: str) -> List[str]:
    with open(file_path, 'r', encoding='utf-8') as dev:
        dataset = [
            row
            for row in dev.read().split('\n')
            if len(row.strip()) > 0
        ]

    return dataset

def save_data(file_path: str, data: List[str]) -> None:
    sorted_data = list(
        sorted(
            filter(lambda a: len(a.strip()) > 0, data),
            key=len,
            reverse=True
        )
    )

    save_document(file_path, '\n'.join(sorted_data))

def append_data(file_path: str, dataset: Union[Set[str], List[str]]) -> int:
    if os.path.exists(file_path):
        for record in load_data(file_path):
            if isinstance(record, set):
                dataset.add(record)
            if isinstance(record, list):
                dataset.append(record)

    dataset = list(sorted(dataset, key=len, reverse=True))

    save_document(file_path, '\n'.join(dataset).strip())

    return len(dataset)



from typing import List, Callable, Iterator, cast
from extr import Entity, TokenGroup
from extr.entities import EntityExtractor
from extr.tokenizers import word_tokenizer
from ..models import Label


class Labeler():
    def __init__(self, tokenizer: Callable[[str], List[str]]) -> None:
        self._tokenizer = tokenizer

    def label(self, text: str, entities: List[Entity]) -> Label:
        return self.__merge(
            word_tokenizer(text, self._tokenizer(text)),
            entities
        )

    def __merge(self, token_group: TokenGroup, entities: List[Entity]) -> Label:
        tokens = list(enumerate(token_group.tokens))
        labels = ['O' for _ in range(len(tokens))]
        for entity in cast(Iterator[Entity], filter(token_group.contains, entities)):
            used_counter = 0

            for i, token in tokens:
                if not entity.is_in(token) and not entity.contains(token):
                    continue

                current_label = labels[i]
                assert current_label == 'O', f'bad tokenizer? multiple entities belong to the same token - {current_label}'

                labels[i] = 'B-' if used_counter == 0 else 'I-'
                labels[i] += entity.label

                token.add_entity(entity)

                used_counter += 1

        return Label(token_group, labels)

class IOB():
    def __init__(self, \
                 tokenizer: Callable[[str], List[str]], \
                 entity_extractor: EntityExtractor) -> None:
        self._labeler = Labeler(tokenizer)
        self._entity_extractor = entity_extractor

    def label(self, text: str) -> Label:
        entities = self._entity_extractor.get_entities(text)
        return self._labeler.label(text, entities)



from typing import Dict, Set, cast
import os

from extr.entities import create_entity_extractor, \
                          LabelOnlyEntityAnnotator, \
                          HtmlEntityAnnotator, \
                          EntityExtractor

from .models import Annotations
from .dev import create_html_file, create_parsed_by_file, create_redacted_file
from ..workspace import load_config, WORKSPACE
from ...utils.filesystem import load_data, save_data
from ...utils import imports


entity_annotator = LabelOnlyEntityAnnotator()
entity_html_annotator = HtmlEntityAnnotator()

def get_extractor() -> EntityExtractor:
    labels = imports.load_file(
        'labels',
        os.path.join(WORKSPACE, 'labels.py')
    )

    return create_entity_extractor(labels.entity_patterns, labels.kb)

def annotate_file(file_path: str) -> Annotations:
    utils = imports.load_file(
        'utils',
        os.path.join(WORKSPACE, 'utils.py')
    )

    entity_extractor = get_extractor()

    cache = Annotations()
    text_by_label: Dict[str, Set[str]] = {}
    for row in load_data(file_path):
        text = cast(str, utils.transform_text(row))
        entities = entity_extractor.get_entities(text)

        cache.text.append(
            entity_annotator.annotate(text, entities)
        )

        cache.html.append(
            entity_html_annotator.annotate(text, entities)
        )

        for entity in entities:
            if not entity.label in text_by_label:
                text_by_label[entity.label] = set()

            text_by_label[entity.label].add(entity.text)

    for key, value in text_by_label.items():
        cache.text_by_label[key] = list(sorted(value))

    return cache

def annotate() -> None:
    cache = annotate_file(
        os.path.join(WORKSPACE, '2', 'dev.txt')
    )

    save_data(
        os.path.join(WORKSPACE, '3', 'dev-ents.txt'),
        cache.text
    )

    create_redacted_file(cache.text)
    create_parsed_by_file(cache.text_by_label)

    config = load_config()
    if bool(config['annotations']['enable-html']):
        create_html_file(cache.html)
